function   model_solution = find_solution( parameters,  filename,  print_level,    initialization  )
% Usage: find_solution( alpha0, beta,  Q,  lambda,  filename  ).  V4.0, using updated market clearing equation.
% 
% This function finds a solution under parameter values alpha0, beta and
% function Q. The solution is saved to file "filename".  This is for the
% model with only w dimension.

% print_level.  0: nothing.  1: important outputs.


%% SECTION 1. Setting PARAMETERS
% (1) Bank Run Parameters and Liquidity Parameteres
% Bank run parameters. Other parameters are inside "find_solution.m".
lambda=parameters.lambda;   alpha0  = parameters.alpha0 ;   beta = parameters.beta;   e = parameters.e;  theta=parameters.theta;  % Bank Run parameters.
Q = parameters.Q ; 
v_prime_fun = parameters.v_prime_fun;
PI = parameters.PI;   % The fraction resellable during a fire sale for the liquid assets.

% (2) Other model parameters
N_w    = parameters.N_w;   max_it=parameters.max_it;  
AL=  parameters.AL;  AH=parameters.AH;
rho = parameters.rho;   sigmaK=parameters.sigmaK;  delta=parameters.delta;    chi=parameters.chi;
investment =  parameters.investment;
investmentD = parameters.investmentD;
muK_fun = parameters.muK_fun;

% (3) Limit values
p0_norun = fsolve(  @(p)  investment(p) + rho*p -  AL,  1 ,   optimoptions( 'fsolve','Display','none' ) );  % The initial condition
p1_norun = fsolve(  @(p)  investment(p) + rho*p -  AH,  1 ,   optimoptions( 'fsolve','Display','none' ) );  % The initial condition
d0 = (AH-AL)/( investmentD(p0_norun) + rho ) * (  (AH-AL)/( p0_norun*sigmaK )^2 * p0_norun + 1  );
psi_w0 = (investmentD(p0_norun) + rho)/ (AH-AL) * d0; 
max_jump= (p1_norun-p0_norun)/p1_norun ;    

% (4)  Grid
gen_grid = @(lower, upper, number)  exp(  linspace( log(lower), log(upper),  number )' ) ;
w_grid = [ 0;  gen_grid(0.001, 0.2,  round(N_w/2) );  gen_grid( 0.21, 1,  N_w-round(N_w/2)-1 )  ];

% (5) Record parameters for call-back inside functions.   par is a struct that includes all the relevant parameters for other functions
par.N    =  N_w;  par.AH  = AH;  par.AL  = AL;  par.rho  = rho;  par.sigmaK  = sigmaK;  par.investment = investment; par.Q = Q;   
par.beta = beta;  par.alpha0=alpha0;   par.investmentD=investmentD;   par.muK_fun = muK_fun;   par.lambda=lambda;
par.e=e;  par.v_prime_fun=v_prime_fun;   par.theta=theta;    par.max_jump=max_jump;
par.p0=p0_norun;   par.p1=p1_norun;    par.d0=d0;   par.psi_w0= psi_w0;


%% SECTION 2. Solution without jump
% Initialize
p_psi1 = zeros( N_w,1 );  p_der_psi1 = zeros(N_w,1);
w_finer_grid = linspace( 0, 1, 200 )';
Qprime_vec = interp1( w_finer_grid, derivative(w_finer_grid, Q(w_finer_grid)),  w_grid ) ;
for(i = 1:(N_w))
    p_psi1(i) = fsolve(  @(p)  investment(p) + rho*p  - AH ,  1 ,   optimoptions( 'fsolve','Display','none' ) );
    p_der_psi1(i) = - rho.*Qprime_vec(i) / investmentD( p_psi1(i) );
end
p_psi1_fitted = interp1(w_grid,p_psi1,'spline','pp');   par.p_psi1_fitted=p_psi1_fitted;
p_der_psi1_fitted = interp1(w_grid(5:N_w),p_der_psi1(5:N_w),'spline','pp');  par.p_der_psi1_fitted = p_der_psi1_fitted; 
% Set options for fsolve:
opts=odeset('Events',@(w,p)  stop_condition_ODE(w,p,par),  'RelTol', 1e-4, 'AbsTol',1e-12);    %  'OutputFcn',@odeplot,
[w, p_sol] = ode45( @(w,p) ODE_noJump(w,p,par) ,  w_grid  ,   p0_norun , opts);
% plot(derivative( w, p_sol ) )
p_SLM = slmengine( w,  p_sol  ,'plot', 'on', 'increasing', 'on',  'concavedown', 'on', 'leftvalue',p0_norun, 'rightvalue',  p1_norun  , 'knots',   floor( numel(p_sol) /4)   );  
p_ode45 = [  slmeval( w_grid(w_grid<=max(w)),  p_SLM);  p1_norun*ones( sum(w_grid>max(w)),1 )    ];
p_SLM = slmengine( w_grid,  p_ode45  ,'plot', 'on', 'increasing', 'on',  'concavedown', 'on', 'leftvalue',p0_norun, 'rightvalue',  p1_norun  , 'knots',   floor( numel(p_sol)/8)   );  

if( nargin==5 )   % If the input does have an initialization
    p_init = initialization.p_new;
    hat_Q_init = initialization.hat_Q_vec;
else
    p_init = slmeval( w_grid,  p_SLM);
    hat_Q_init = 0.05*Q(w_grid);
end

%% SECTION 3. Full Solution. With a different methods that iterate over p() and x^g()
tol = 0.001;     learning_rate = 0.6;  % Note: 60% picking the newly updated values, but 40% keeping the old ones. 
% Initialization
p_new = p_init;   p_last  = p_init;  
xg_new = Q(w_grid)./( 2 * w_grid .* p_last ) ;  xg_new(1) = 2*xg_new(2); xg_new(end) =e; xg_last = xg_new + 1;
hat_Q_new = hat_Q_init;  hat_Q_last = hat_Q_new;    hat_Q_stablizer = false;
counter_aggregate = 0;  errors = [];  tic;   error_current=10;
% Main Loop
while( counter_aggregate<=max_it  &&  error_current>tol )
    if( counter_aggregate>5 && error_current>2*errors(end-1))   % At least do five iterations. If the error increased by more than 2 times, just stop.
        break;
    end
    close all;      counter_aggregate = counter_aggregate+1;       
    disp( [ '******************   solving the model with Q/GDP= ', num2str(100*Q(1)/0.145,3), '%:  iteration ', num2str(counter_aggregate), '   ******************' ] );
    if( counter_aggregate>2 )
        disp( [ 'Current error = ', num2str(error_current,3), ', change by ',  num2str( 100*(log(error_current)-log(errors(end-1))), 3 ), '%. e(hat_Q)=', num2str( sum(abs(hat_Q_new-hat_Q_last)), 3), ', e(p)=', num2str( sum(abs(p_new-p_last)), 3)  ] )
    end
    % Step 1: Update for a new round.
    p_last = p_new;     xg_last = xg_new;   hat_Q_last = hat_Q_new; % A new round of solution. Only price function and kappap are denoted by _new and _last. All others are denoted by _vec.
    pw_vec = derivative( w_grid, p_last );   
    p_fitted = interp1( w_grid, p_last, 'linear', 'pp' );

    psi_vec =  ( investment( p_last ) + rho * (p_last+hat_Q_last)   - AL ) / (AH-AL) ;
    xK_vec = psi_vec ./ w_grid  .* p_last ./ ( p_last + hat_Q_last ) ;  

    kappap_vec = zeros( size(w_grid) );   % Store the jump results.
    solved_kappap = ones( size(w_grid) );   solved_xK = ones( size(w_grid) );   solved_xg = ones( size(w_grid) );     
    solved =  ones( size(w_grid) ); % Whether the solutions exist  

    % Step 2: Solve p and xg for each grid point.
    for( i = 2:(numel(w_grid)-1) )  
        w = w_grid(i); p=p_last(i);   xK = xK_vec(i);   xg=xg_last(i);  pw=pw_vec(i);   hat_Q = hat_Q_last(i);  par.hat_Q=hat_Q;
        [ xK, xg, kappap, exit_flag_xK, exit_flag_xg, exit_flag_kappap ] = inside_iteration( w, p, p_fitted, xK, xg, pw, i, print_level, par );
        %Update the new price function.
        psi = xK*w ./ (  p/(p+hat_Q)  );
        p = fsolve(  @(p)  investment(p) + rho*(p+hat_Q) -  (psi*AH+(1-psi)*AL) ,  (p0_norun+p1_norun)/2  ,   optimoptions( 'fsolve','Display','none' ) );  % The initial condition
        p_new(i) = p;   xg_new(i) = xg;  kappap_vec(i) =  kappap;
        solved_xK(i) = exit_flag_xK;   solved_xg(i) = exit_flag_xg;    solved_kappap(i) = exit_flag_kappap;   
    end  % End of individual point solution

    solved_index = find((solved_xK==1&solved_xg==1&solved_kappap==1)&(w_grid<1));
    disp( [ num2str(100*(numel(solved_index)+1)/numel(w_grid)), '% is solved' ] )

    % Post processing and interpolation
    p_left = p_new(solved_index(1));  p_right=p_new(solved_index(end));
    p_SLM = slmengine( w_grid(solved_index),  p_new(solved_index) ,    'increasing', 'on',  'leftvalue',p_left, 'rightvalue',  p_right , 'concavedown', 'on',  'knots', round(numel(w_grid(solved_index))/2)   );   pause(1);
    p_new = slmeval( w_grid ,  p_SLM);

    xg_new = interp1( w_grid(solved_xg==1), xg_new(solved_xg==1), w_grid );
    kappap_vec = interp1( w_grid(solved_kappap==1), kappap_vec(solved_kappap==1), w_grid,  'linear', 'extrap' );
    

    % Step 3: Update hat_Q using simulation methods. (Important: I need to update the hat_Q in each round accurately. Try a better method. )
    model_solution.p_ode45 = p_ode45; 
    model_solution.w_grid = w_grid; 
    model_solution.p_new = p_new;  model_solution.p0 = p_left;   model_solution.p1 = p_right;  
    model_solution.xg_vec = xg_new; 
    model_solution.kappap_vec = kappap_vec;
    model_solution.parameters = parameters;
    model_solution.errors = errors;
    model_solution.hat_Q_vec = hat_Q_last;
    S = post_processing(model_solution);
    
    if( print_level>=1 )
        figure('position', [0, 0, 1000, 600]);
        subplot(2,2,1); plot( w_grid(solved_index),   kappap_vec(solved_index) ); xlabel('$w$', 'Interpreter','latex'); ylabel('$\kappa^p$', 'Interpreter','latex');   title('Price jump');
        subplot(2,2,2); plot( w_grid(solved_index), p_new(solved_index), '--',  w_grid(solved_index), p_last(solved_index),'-.' ); xlabel('$w$', 'Interpreter','latex');  ylabel('$p$', 'Interpreter','latex');  title('Capital price');  %legend('new', 'last', 'Location', 'northwest');
        psi_vec =  ( investment( p_new ) + rho * p_new  - AL ) / (AH-AL) ;
        frac = xg_new.*w_grid./(Q(w_grid)./(p_new+hat_Q_new));
        subplot(2,2,3); plot( w_grid(solved_index), S.liq_prem_vec(solved_index) ); xlabel('$w$', 'Interpreter','latex');   ylabel('liquidity premium');  title('liquidity premium');
        subplot(2,2,4); plot( w_grid(solved_index), S.muw_vec(solved_index) ); xlabel('$w$', 'Interpreter','latex');   ylabel('muw');  title('growth of banker wealth share');
        pause(2);
    end
    
    if( hat_Q_stablizer==true )
        hat_Q_new = hat_Q_last;
    else
        disp( 'Evaluate hat_Q...' )
        hat_Q_update = hat_Q_calculation(S);
        hat_Q_diff = sum( abs( hat_Q_update -  hat_Q_last) );
        if( (hat_Q_diff < 0.001 && Q(1)/0.145<=1) || (hat_Q_diff < 0.01 && Q(1)/0.145>1) )
            hat_Q_stablizer = true;   disp( 'hat Q stabalizer started!!!  :)' ) ;        
        end
        hat_Q_new = hat_Q_update; 
    end
    evaluation_grid = linspace(0,1,100)';
    disp(  [ 'Current Q_hat/Q ratio = ',  num2str(   mean( interp1( w_grid, hat_Q_new, evaluation_grid ) ) / mean(Q(evaluation_grid)) , 3 ) ]  )
    if( print_level>=1 )
        figure(2);  plot(w_grid, hat_Q_new );  ylabel('hat Q');  
        pause(2);
    end

    % Step 4: Use the learning rate to update new values.
    % IMPORTANT: Update according to the learning rate!
    p_new = learning_rate*p_new + (1-learning_rate)*p_last;
    xg_new = learning_rate*xg_new + (1-learning_rate)*xg_last;
    hat_Q_new =  learning_rate*hat_Q_new + (1-learning_rate)*hat_Q_last;
    toc   % Show time
    error_current=  sum( abs(p_new-p_last) + abs(hat_Q_new-hat_Q_last) );  errors = [ errors, error_current ];
end  % End of whole while loop
p_new=  p_last;    hat_Q_new = hat_Q_calculation(S);
disp(  [ '*** Final Round: sum of absolute error = ' , num2str( error_current, 3 ) ] )
close all;

%% Section 5. Model Output
model_solution.p_ode45 = p_ode45; 
model_solution.w_grid = w_grid; 
model_solution.p_new = p_new;  model_solution.p0 = p_new(1);   model_solution.p1 = p_new(end);  
model_solution.xg_vec = xg_new; 
model_solution.kappap_vec = kappap_vec;
model_solution.parameters = parameters;
model_solution.errors = errors;
model_solution.hat_Q_vec = hat_Q_new;

if( filename~="" )
    save(filename, '-struct', 'model_solution');
end



